/*
 * ObservableSolver.java
 * Created on 21.04.2003
 */
package search.graph;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Observable;
import java.util.List;

import search.util.Queue;

/**
 * Die Klasse vereint einen Graphsuche-Algorithmus und die noetigen
 * Methoden um seinen Ablauf zu beobachten.
 * <p>
 * Die Beobachtung wird durch Verwendung des Beobachter-Musters gemaess
 * <code>java.util.Observable</code> und <code>java.util.Observer</code>
 * unterstuetzt.
 * <p>
 * Das Loesungsverhalten des Algorithmus haengt entscheident von der von der
 * Suchstrategie bereitgestellten Datenstruktur
 * (<code>search.util.Queue</code>) ab. Sofern,
 * diese keine Elemente "vergisst" oder "verdoppelt", kann aber garantiert
 * werden, dass bei einem endlichen Graphen eine Loesung gefunden wird, sofern
 * sie existiert oder dass anderfalls in endlicher Zeit die Nichtexistenz einer
 * Loesung festgestellt wird. Bei unendlichen Graphen sind keine allgemeinen
 * Aussagen moeglich. Die Art und Qualitaet der Loesung haengt ausschliesslich
 * von dem Verhalten des Verfahrens ab. 
 * <p>
 * Da in der Regel der Algorithmus nebenlaeufig ausgefuehrt wird, ist
 * er als eine <code>run</code>-Methode realisiert.
 * Die Klasse ist insofern Thread-sicher als sie
 * jederzeit unterbrochen werden kann, und insofern als die
 * Verzoegerungszeit jederzeit veraendert werden kann.
 * Die Suche wartet nicht auf die graphische Darstellung durch evtl. Beobachtern.
 * <p>
 * Die Unterbrechung per <code>Thread.interrupt()</code> ist nicht vorgesehen
 * und wird als fataler Fehler (<code>Error</code>) behandelt.
 * 
 * @author Erich Ehses.
 * 
 * @see java.util.Observer
 * @see java.util.Observable
 * @see java.lang.Runnable
 */
public class ObservableSolver extends Observable implements Runnable {

    private Queue openQueue; // Liste der zu verfolgenden Pfade.
    private Path path; // aktueller Pfad
    private boolean running; // Zustand der Ausfuehrung
    private volatile long delay = 1; // Zeitverzoegerung pro Schritt
    private SearchStrategy searchStrategy;

    private int timeToLive;
    private double explorationProbability;

    /**
     * Konstruktor.
     */
    public ObservableSolver() {
    }

    /**
     * Legt die Suchparameter fest.
     * @param strategy bestimmt das Suchverfahren.
     * @param startNode Name des Startknotens.
     * @param goalNode Name des Zielknotens.
     */
    public synchronized void start(SearchStrategy strategy,
        Node startNode, Node goalNode)
    {
        searchStrategy = strategy;
        while (running)
            try {
                wait();
            }
            catch (InterruptedException e) {
                reportFatalError();
            }
        openQueue = strategy.makeQueue();
        path = new Path(startNode, goalNode);
        running = true;
        notify(); // es gibt maximal 1 wait()
    }

    private void reportFatalError() {
        throw new Error("InterruptedException should not occur!");
    }

    /**
     * Veraendert die Verzoegerung pro Schritt.
     * Bei dem Wert 0 findet kein Warten statt und in diesem
     * Fall erfolgt weder eine Verlaufsmeldung noch kann die Suche
     * abgebrochen werden.
     * @param delay Verzoegerung in Millisekunden.
     */
    public synchronized void setDelay(long delay) {
        if (delay < 0)
            delay = 0;
        this.delay = delay;
        notify(); // unterbreche die laufende Verzoegerung
    }

    /**
     * Haelt die Suche an.
     */
    public synchronized void stop() {
        running = false;
        notify(); // es gibt maximal 1 wait()
    }

    /**
     * Fragt ab, ob die Suche gerade ausgefuehrt wird.
     * @return <code>true</code>, wenn die Suche laeuft.
     */
    public synchronized boolean isRunning() {
        return running;
    }

    /**
     * Endlosschleife zur Durchfuehrung von Suchauftraegen.
     * Die Suche wird mit <code>start</code> gestartet.
     * Dabei werden auch die Loesungsparameter definiert.
     * <p>Wenn die Verzoegerungszeit groesser als 0 ist, werden
     * alle registrierten Beobachter von Veraenderungen informiert.
     * Dabei wird die Methode <code>update</code> aufgerufen. Ihr
     * wird als zweiter Parameter das aktuelle Path-Objekt
     * uebergeben.
     * Die Suche wird durch (nebenlaeufigen) Aufruf der
     * Methode <code>stop</code> abgebrochen.
     * 
     * @see ObservableSolver#setDelay
     * @see ObservableSolver#stop
     */
    public void run() {
        try {
            // closedSet speichert die Menge der besuchten Knoten.
            HashSet closedSet = new HashSet();

            while (true) {
                // warte auf den Startschuss:
                synchronized (this) {
                    while (!running) wait();
                    // der (mit setParams) festgelegte Startpfad wird gespeichert.
                    openQueue.clear();
                    openQueue.put(path);
                }
                // es wurden noch keine Knoten besucht.
                closedSet.clear();
                // informiere alle Beobachter.
                setChanged();
                notifyObservers(path);

                // normale Suche
                if (!searchStrategy.equals(SearchStrategy.ANT_NET)){
                    // Suchschleife normal
                    while (isRunning() && !openQueue.isEmpty()) {
                        // entnehme Pfad #1
                        path = (Path)openQueue.get();
                        // fertig?
                        if (path.isComplete()) break;
                        // Jetzt mssen wir ab dem Front-Knoten weitersuchen.
                        Node front = path.getLastNode();
                        // schon behandelte Knoten werden ignoriert.
                        if (!closedSet.contains(front)) {
                            // dieser Knoten ist abgeschlossen.
                            closedSet.add(front);
                            // wir muessen die Pfade zu den noch nicht
                            // besuchten Nachbarn vormerken.
                            for (Iterator itr=front.neighbours(); itr.hasNext();) {
                                Node location = (Node)itr.next();
                                if (!closedSet.contains(location)) {
                                    openQueue.put(path.expandWith(location));
                                }
                            }
                            // informiere alle Beobachter.
                            if (delay != 0) {
                                setChanged();
                                notifyObservers(path);
                                // sitze die Verzoegerungszeit ab.
                                synchronized (this) {
                                    wait(delay);
                                }
                            }
                        }
                    }
                }
                // Ameisensuche
                else{
                    while (isRunning()){
                        // kompletter Weg gefunden
                        if (path.isComplete()) {
                            //System.out.println("!!!!! kompletten Weg gefunden (Lnge : "+path.getPathLength()+" km)");
                            break;
                        }
                        // Time to Live berschritten
                        if (path.getNodes().size() > timeToLive){
                            System.out.println("!!!!! Abbruch wegen TTL-berschreitung");
                            break;
                        }
                        //letzter bereits zum Pfad gehrender Knoten
                        Node front = path.getLastNode();

                        // Die Entscheidung, ob der beste bekannte Nachbarknoten oder ein zuflliger gewhlt wird
                        double decision = Math.random();

                        // falls sich der Zielknoten in direkter Nachbarschaft befindet wird auf jeden Fall dieser gewhlt
                        boolean goalIsDirectNeighbour=false;
                        Iterator it = front.neighbours();
                        Node temp;
                        while (it.hasNext()){
                            temp = (Node)it.next();
                            if (temp.equals(path.getGoalNode())){
                                path = path.expandWith(temp);
                                goalIsDirectNeighbour=true;
                                break;
                            }
                        }
                        // anderenfalls den bestmglichen Weg gehen es bereits einen oder mehrere Wege zu diesem Ziel gibt
                        if (decision > explorationProbability && null != front.getNeighbourNodeWithShortestPath(path.getGoalNode()) && !goalIsDirectNeighbour){
                            //System.out.println("***** bestmglichen Weg gehen");
                            temp = front.getNeighbourNodeWithShortestPath(path.getGoalNode());
                            path = path.expandWith(temp);

                        }
                        // einen zuflligen Weg gehen
                        else {
                            if (!goalIsDirectNeighbour){
                                temp = front.getRandomNeighbourNode();
                                path = path.expandWith(temp);

                            }
                        }
                        
                        // informiere alle Beobachter.
                        if (delay != 0) {
                            setChanged();
                            notifyObservers(path);
                            // sitze die Verzoegerungszeit ab.
                            synchronized (this) {
                                wait(delay);
                            }
                        }
                    }
                }
                if (searchStrategy.equals(SearchStrategy.ANT_NET) && path.isComplete()){
                    //Ausgabe des gegangenen Weges
                    //System.out.println("");
                    //System.out.println("(forward Ant) gegangener Weg: ");
                    //System.out.println("");
                    List list = path.getNodes();
                    for (int i =list.size()-1; i>-1; i--){
                        Node node = (Node) list.get(i);
                        //System.out.print(" -> "+node.getName());
                    }
                    //System.out.println("");
                        //Pheromonwerte erhhen
                        for (int i=0; i<list.size(); i++){
                            if (i>0){
                                Node aktuell = (Node) list.get(i);
                                Node lastVisitedNeighbourNode = (Node) list.get(i-1);
                                Node goalNode = path.getGoalNode();
                                aktuell.increasePheromoneValue(goalNode, lastVisitedNeighbourNode, path);
                            }
                        }

                }

                // fertig oder abgebrochen.
                stop();
                // informiere alle Beobachter.
                setChanged();
                notifyObservers(path);
            }
        }
        catch (InterruptedException e) {
            reportFatalError();
        }
    }
    public double getExplorationProbability() {
        return explorationProbability;
    }

    public void setExplorationProbability(double explorationProbability) {
        this.explorationProbability = explorationProbability;
    }

    public int getTimeToLive() {
        return timeToLive;
    }

    public void setTimeToLive(int timeToLive) {
        this.timeToLive = timeToLive;
    }
}
