/*
 * Path.java
 * Created on 21.04.2003
 */
package search.graph;

import java.util.List;
import java.util.Iterator;
import java.util.Vector;

import search.util.SharedListTree;
import search.util.TreeNode;

/**
 * Speichert einen Weg in einer Knotenfolge.
 * Waehrend der Ausfuehrung der Suche ist die Knotenfolge umgekehrt zur
 * Wegrichtung.
 * @author Erich Ehses.
 */
public final class Path {
    private double pathLength = 0.0;    // Laenge des Weges.
    private SharedListTree nodes = null;          // Liste der Knoten.
    private Node goalNode;
    private Node startNode;

    /**
     * Implementierungshilfsmittel fuer <code>expandWith</code>.
     */
    private Path() {}
    
    /**
     * Konstruktor.
     * Zwecks Visualisierung wird gleichzeitig mit der Pfad-Liste
     * ein Suchbaum aufgebaut.
     * @param startNode Startknoten.
     * @param goalNode Zielknoten
     *  Zielknoten.
     */
    public Path(Node startNode, Node goalNode) {
        nodes = new SharedListTree();
        nodes.add(0, startNode);
        this.goalNode = goalNode;
        this.startNode = startNode;
    }
    
    
    /**
     * Gibt den Endknoten des Pfades zurueck.
     * @return Endknoten.
     */
    public Node getLastNode() {
        return (Node)nodes.get(0);
    }

    /**
     * Gibt den aktuellen Knoten des Pfades zurueck.
     * (fr die Ameisensuche)
     * @return Endknoten.
     */
    public Node getCurrentNode() {
        return (Node)nodes.get(1);
    }

    /**
     * Gibt den Zielknoten der Pfadsuche zurueck.
     * @return Zielknoten.
     */
    public Node getGoalNode() {
        return goalNode;
    }
    
    /**
     * Untersucht ob der Pfad bis zum Ziel reicht.
     * @return <code>true</code> wenn das Ziel gefunden wurde.
     */
    public boolean isComplete() {
        return goalNode == nodes.get(0);
    }

    /**
     * Erweitert den Path um einen weiteren Knoten.
     * @param nextNode neuer Knoten
     * @return neuer erweiterter Pfad.
     */
    public Path expandWith(Node nextNode) {
        if (!nextNode.isDead){
            // wurde der Knoten bereits besucht wird nur der bis hier gegangene Weg behalten (der Kreis wird komplett vergessen)
            if (nodes.contains(nextNode)){
                //System.out.println("!!!!! Im Kreis gegangen ("+nextNode.getName()+")");
                Path p = new Path(startNode, goalNode);
                List list = getNodes();
                for (int i =list.size()-2; i>-1; i--){
                    Node node = (Node) list.get(i);
                    //System.out.print(node.getName()+", ");
                    if (!node.equals(nextNode)  || node.equals(p.startNode)){
                        p.pathLength = p.pathLength + nextNode.edgeTo(getLastNode()).getEdgeLength();
                        p.nodes.add(0,node);
                    }
                    else{
                        break;
                    }
                }
                //System.out.println("Start: "+startNode.getName()+" Ziel: "+goalNode.getName());
                return p;
            }
            else{
            Path p = new Path();
            //System.out.println("Alles Gut! ("+nextNode.getName()+")");
            p.pathLength = pathLength + nextNode.edgeTo(getLastNode()).getEdgeLength();
            p.nodes = new SharedListTree(nodes);
            p.nodes.add(0, nextNode);
            p.goalNode = goalNode;
            p.startNode = startNode;
            return p;
            }
        }
        else {
            return this;
        }
    }
    
    /**
     * Gibt die Pfadlaenge zurueck.
     * @return Pfadlaenge.
     */
    public double getPathLength() {
        return pathLength;
    }

    //ermittelt die Strecke von einem Knoten des Pfades zum Zielknoten
    public double getLengthFrom (Node from){
        double length=0;

        // umschichten
        Iterator itBase = nodes.iterator();
        Vector vec1 = new Vector();
        Vector vec2 = new Vector();
        while (itBase.hasNext()){
            vec1.add(itBase.next());
        }
        for (int i=vec1.size()-1; i>=0;i--){
            vec2.add(vec1.elementAt(i));
        }
        Iterator it=vec2.iterator();

        Node actual=null;
        Node next=null;
        boolean add=false;
        if (it.hasNext()){
            actual =(Node) it.next();
        }
        //System.out.println("*** neue Lngenberechnung");
        while (it.hasNext()){

            next = (Node)it.next();
            if (null != actual && null != next){
                //ab jetzt muss addiert werden
                if (actual.equals(from)){
                    add = true;
                }
            }

            if (add){
                length += actual.edgeTo(next).getEdgeLength();
                //System.out.println("actual : "+actual.getName()+", next : "+next.getName()+" , distance : "+actual.edgeTo(next).getEdgeLength());
            }
            actual = next;
        }
        //System.out.println("*** Lngenberechnung zuende (length :"+length+" )");
        return length;
    }

    /**
     * Gibt die Liste aller Pfadknoten zurueck.
     * Die Reihenfolge der Knoten ist umgekehrt zur Pfadrichtung.
     * @return Liste der Knoten.
     */
    public List getNodes() {
        return nodes;
    }
    
    /**
     * Gibt den Startknoten des Pfades zurck.
     * Von diesem Startknoten aus kann der gesamte Suchbaum untersucht
     * werden.
     * @return Startknoten.
     */
    public TreeNode getRoot() {
        return nodes.getRoot();
    }
    
    /**
     * Gibt die Anzahl der Knoten zurueck.
     * Dies ist gleich der graphtheoretischen Pfadlaenge + 1.
     * @return Anzahl der Knoten.
     */
    public int size() {
        return nodes.size();
    }
}
