/*
 * SearchStrategy.java
 * Created on 21.04.2003
 */
package search.graph;

import java.util.Comparator;

import search.util.*;

/**
 * Die Klasse stellt ber konstante Referenzen Factory-Objecte fr
 * Queue-Datenstrukturen bereit. Das von der jeweiligen Konstanten
 * mittels <code>makeQueue</code> erzeugte <code>Queue</code>
 * Objekt dient zur Steuerung des Graphsuchalgorithmus.
 * <p>
 * Beispiele:
 * <pre>
 *     Queue q = SearchStrategy.FIFO.makeQueue();
 * 
 *     Path p = Solver.solve(SearchStrategy.A_STAR, start, goal);
 * </pre> 
 * 
 * @author Erich Ehses.
 */
public abstract class SearchStrategy {
    private SearchStrategy() {}
    
    /**
     * Erzeugt eine Datenstruktur zur Steuerung der Graphsuche.
     * @return Queue-Objekt mit geeigneter Ordnung.
     */
    public abstract Queue makeQueue();

    /**
     * Tiefensuche.
     * Auswahl nach dem LIFO-Prinzip.
     */
    public static final SearchStrategy LIFO =
        new SearchStrategy() {
            public Queue makeQueue() {
                return new LIFOQueue();
            }
        };
     
     /**
      * Breitensuche.
      * Auswahl nach dem FIFO-Prinzip.
      */
     public static final SearchStrategy FIFO =
         new SearchStrategy() {
             public Queue makeQueue() {
                 return new FIFOQueue();
             }
         };
     
     /**
      * Dijkstra's Algorithmus.
      * Es wird jeweils der Knoten mit dem krzesten Weg vom
      * Startpunkt ausgewhlt.
      */
     public static final SearchStrategy PATH_LENGTH =
         new SearchStrategy() {
             public Queue makeQueue() {
                 return new PriorityQueue(new PathLengthComparator());
             }
         };
     
     /**
      * Hill-Climbing. Es wird jeweils der dem Ziel am nchsten liegende
      * Knoten ausgewhlt.
      */
     public static final SearchStrategy HILL_CLIMBING =
        new SearchStrategy() {
            public Queue makeQueue() {
                return new PriorityQueue(new HillClimbingComparator());
            }
        };
        
     /**
      * A*-Suche.
      * Es wird jeweils der Knoten ausgewhlt, fr den die Summe von
      * zurckgelegtem Weg und Entfernung zum Ziel am kleinsten ist.
      */
     public static final SearchStrategy A_STAR =
        new SearchStrategy() {
            public Queue makeQueue() {
                return new PriorityQueue(new AStarComparator());
            }
        };

    /**
      * Ant-Net Algorithmus.
      * Es wird jeweils der Knoten mit dem hchsten Pheromonwert fr das
      * gewhlte Ziel  oder ein zuflliger anderer Knoten ausgesucht (Wahrscheinlichkeit einstellbar)
      */
     public static final SearchStrategy ANT_NET =
         new SearchStrategy() {
             public Queue makeQueue() {
                 return new AntQueue(new PheromoneValueComparator());
             }
         };

    /**
     * Vergleicht die Pheromonkonzentrationen
     */
    private static class PheromoneValueComparator implements Comparator {
        private Node goalNode = null;
        private Node neighbourNode1 = null;
        private Node neighbourNode2 = null;
        private Node currentNode = null;

        /* (non-Javadoc)
         * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
         */
        public int compare(Object o1, Object o2) {
            Path p1 = (Path)o1;
            Path p2 = (Path)o2;
            goalNode = p1.getGoalNode();
            neighbourNode1 = p1.getLastNode();
            neighbourNode2 = p2.getLastNode();
            currentNode = p1.getCurrentNode();

            //double dist = p1.getPathLength() - p2.getPathLength();
            int pheromoneValue = currentNode.getPheromoneValue(neighbourNode1, goalNode) - currentNode.getPheromoneValue(neighbourNode2, goalNode);
            //System.out.println("***** Pheromonkonzentration von "+neighbourNode1.getName()+" -> "+goalNode.getName()+" : "+currentNode.getPheromoneValue(neighbourNode1, goalNode));
            //System.out.println("***** Pheromonkonzentration von "+neighbourNode2.getName()+" -> "+goalNode.getName()+" : "+currentNode.getPheromoneValue(neighbourNode2, goalNode));
            //System.out.println("");
            return (pheromoneValue < 0)? -1: (pheromoneValue == 0)? 0: +1;
        }
    }

    /**
     * Vergleicht die zurueckgelegten Wege.
     */
    private static class PathLengthComparator implements Comparator {
        /* (non-Javadoc)
         * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
         */
        public int compare(Object o1, Object o2) {
            Path p1 = (Path)o1;
            Path p2 = (Path)o2;
            double dist = p1.getPathLength() - p2.getPathLength();
            return (dist < 0)? -1: (dist == 0)? 0: +1;  
        }
    }
    
    /**
     * Vergleicht die zurueckgelegten Wege + Luftlinie zum Ziel.
     */
    private static class AStarComparator implements Comparator {
        private Node goal = null;
        /* (non-Javadoc)
         * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
         */
        public int compare(Object o1, Object o2) {
            Path p1 = (Path)o1;
            Path p2 = (Path)o2;
            if (goal == null) goal = p1.getGoalNode();
            double dist = p1.getPathLength() - p2.getPathLength()
                + p1.getLastNode().distTo(goal)
                - p2.getLastNode().distTo(goal);
            return (dist < 0)? -1: (dist == 0)? 0: +1;  
        }
    }
    
    /**
     * Vergleicht die Entfernung zum Ziel.
     */
    private static class HillClimbingComparator implements Comparator {
        private Node goal = null;
        /* (non-Javadoc)
         * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
         */
        public int compare(Object o1, Object o2) {
            Path p1 = (Path)o1;
            Path p2 = (Path)o2;
            if (goal == null) goal = p1.getGoalNode();
            double dist = 
                  p1.getLastNode().distTo(goal)
                - p2.getLastNode().distTo(goal);
            return (dist < 0)? -1: (dist == 0)? 0: +1;  
        }
    }    
}
