/*
 * Solver.java
 * Created on 21.04.2003
 */
package search.graph;

import java.util.HashSet;
import java.util.Iterator;

import search.util.Queue;

/**
 * Bietet eine generalisierte Suchmethode.
 * @author Erich Ehses.
 */
public final class Solver {
    private Solver() {}

    
    /**
     * Fuehrt eine Graphsuche durch.
     * Das zurueckgegebene Objekt enthaelt die Liste der besuchten Ort in
     * umgekehrter Reihenfolge.
     * <p>
     * Das Loesungsverhalten des Algorithmus haengt entscheident von der
     * von der Suchstrategie bereichtgestellten Datenstruktur
     * (<code>search.util.Queue</code>) ab. Sofern,
     * diese keine Elemente "vergisst" oder "verdoppelt", kann aber garantiert
     * werden, dass bei einem endlichen Graphen eine Loesung gefunden wird,
     * sofern sie existiert oder dass anderfalls in endlicher Zeit die 
     * Nichtexistenz einer Loesung festgestellt wird. Bei unendlichen Graphen
     * sind keine allgemeinen Aussagen moeglich. Die Art und Qualitaet der
     * Loesung haengt ausschliesslich von dem Verhalten des Verfahrens ab. 
     * 
     * @param strategy die zu verwendende Suchstrategie.
     * @param startNode Startpunkt.
     * @param goalNode Zielpunkt.
     * @return <code>Path</code> Objekt oder <code>null</code> wenn kein
     *          Weg gefunden wurde.
     */
    public static Path solve(SearchStrategy strategy, 
        Node startNode, Node goalNode)
    {
        // Menge der abgeschlossenen Knoten.                   
        HashSet closedSet = new HashSet();
        // Liste der zu untersuchenden Pfade.
        Queue queue = strategy.makeQueue();
        queue.put(new Path(startNode, goalNode));
        while (!queue.isEmpty()) {
            // verfolge Pfad #1, wenn der Endknoten noch nicht anders
            // erreicht wurde.
            Path path = (Path)queue.get();
            // schon fertig?
            if (path.isComplete()) return path;
            // nein: weitersuchen !
            Node front = path.getLastNode();
            if (!closedSet.contains(front)) {
                // der Knoten ist jetzt abgeschlossen.
                closedSet.add(front);
                // die Pfade zu den Nachbarn muessen vorgemerkt werden.
                for (Iterator iter = front.neighbours(); iter.hasNext();) {
                    Node location = (Node)iter.next();
                    if (!closedSet.contains(location)) {
                        queue.put(path.expandWith(location));
                    }
                }
            }
        }
        // kein Weg gefunden.
        return null;
    }     
}
