/*
 * DataPanel.java
 * Created on 24.04.2003
 */
package search.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.DecimalFormat;
import java.util.Observable;
import java.util.Observer;

import javax.swing.*;
import javax.swing.border.BevelBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import search.graph.*;
import search.util.TreeAlgorithms;

/**
 * Die Klasse <code>DataPanel</code> verwaltet den Benutzerdialog.
 * Von hier aus wird die Graphsuche gestartet. Das Objekt registriert
 * sich als Beobachter der Suche und stellt die aktuelle Weglaenge dar.
 * 
 * @author Erich Ehses.
 */
public final class DataPanel extends JPanel implements Observer {
    
    /*
     * Konstanten fuer das Aussehen der Graphik.
     */
    private static final Color BUTTON_COLOR = Color.darkGray;
    private static final Color START_COLOR = Color.green;
    private static final Color STOP_COLOR = Color.red;
    
    /*
     * Modelldaten.
     */
    private Graph graph;                // der Graph, in dem gesucht wird
    private ObservableSolver solver;    // der Problemloeser

    /*
     * Zahlenformate.
     */
    private DecimalFormat lengthFormat = new DecimalFormat("0.0");
    private DecimalFormat countFormat = new DecimalFormat("0");
    
    /*
     * Widgets.
     */
    private LabelledItemPanel items;
    private JComboBox startSelection;
    private JComboBox goalSelection;
    private JTextField antNetLengthField;
    private JTextField dijkstraLengthField;
    private JTextField relativeLengthField;
    private JTextField timeToLive;
    private JTextField explorationProbability;
    private JTextField antGenerationIntervall;
    private RadioButtonPanel algoSelection;
    private JSlider delaySlider;
    private JButton startButton;
    private JButton stopButton;
    private JCheckBox dijkstraCheckbox;
    private JCheckBox removeCheckbox;

    //soll eine Ameisen oder eine Bestwegsuche laufen
    private boolean antToggle;

    //Name des Start- und Endknotens
    private String startName;
    private String goalName;

    //Die Weglngen
    double dijkstraLength;
    double antNetLength;


    /**
     * Die Namen der Algorithmen.
     */  
    private static final String[] algoNames =
        {
            "Tiefensuche",
            "Breitensuche",
            "Dijkstra",
            "Bergsteigen",
            "A*"
        };

    /**
     * Die Tags der Algorithmen.
     */
    private static final SearchStrategy[] algorithms =
        {
            SearchStrategy.LIFO,
            SearchStrategy.FIFO,
            SearchStrategy.PATH_LENGTH,
            SearchStrategy.HILL_CLIMBING,
            SearchStrategy.A_STAR
        };

    private boolean pleaseContinue;

    /**
     * Konstruktor. Erzeugt ein fertiges Eingabepanel.
     * @param solver das suchende Objekt.
     * @param graph der Graph in dem gesucht wird.
     */
    public DataPanel(ObservableSolver solver, final Graph graph) {
        this.solver = solver;
        this.graph = graph;
        pleaseContinue = true;

        setLayout(new BoxLayout(this,BoxLayout.Y_AXIS));
        //setBorder(new BevelBorder(BevelBorder.RAISED));
        
        initLabelledItemPanel();
        //add(items, BorderLayout.NORTH);
        add(items);

        startButton = new JButton();
        setButtonToStart();
        startButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                startAntApplication();
                pleaseContinue = true;
            }
        });
        //add(startButton, BorderLayout.SOUTH);
        add(startButton);

        stopButton = new JButton();
        stopButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                pleaseContinue = false;
            }
        });
        stopButton.setFont(new Font("SansSerif", Font.BOLD, 18));
        stopButton.setBackground(BUTTON_COLOR);
        stopButton.setForeground(STOP_COLOR);
        stopButton.setText("stop");
        //add(stopButton, BorderLayout.EAST);
        add(stopButton);

        antToggle = true;
    }
    
    /**
     * Initialisiert die Komponenten des Unterfensters
     * fuer die Eingabe.
     */
    //todo hier wird das linke Teilfenster zusammengesetzt
    private void initLabelledItemPanel() {
        items = new LabelledItemPanel();
        
        String[] nodeNames = graph.getNodeNames();
        
        startSelection = new JComboBox(nodeNames);
        //items.addItem("Startpunkt", startSelection);
        
        goalSelection = new JComboBox(nodeNames);
        //items.addItem("Zielpunkt", goalSelection);



        algoSelection = new RadioButtonPanel(algoNames);
        //items.addItem("Algorithmus", algoSelection);


        //todo: hier sind die relevanten Infos
        timeToLive = new JTextField("20");
        timeToLive.setHorizontalAlignment(JTextField.RIGHT);
        timeToLive.setEditable(true);
        items.addItem("Time to live", timeToLive);

        explorationProbability = new JTextField("0.2");
        explorationProbability.setHorizontalAlignment(JTextField.RIGHT);
        explorationProbability.setEditable(true);
        items.addItem("Entdeckungstrieb", explorationProbability);

        antGenerationIntervall = new JTextField("1");
        antGenerationIntervall.setHorizontalAlignment(JTextField.RIGHT);
        antGenerationIntervall.setEditable(true);
        //items.addItem("Ameisengenerierungs-Intervall (Sekunden)", antGenerationIntervall);

        dijkstraCheckbox = new JCheckBox();
        dijkstraCheckbox.setSelected(false);
        items.addItem("parallele Bestwegsuche",dijkstraCheckbox);

        removeCheckbox = new JCheckBox();
        removeCheckbox.setSelected(false);
        //items.addItem("Knoten entfernen",removeCheckbox);

        antNetLengthField = new JTextField("-");
        antNetLengthField.setHorizontalAlignment(JTextField.RIGHT);
        antNetLengthField.setEditable(false);
        items.addItem("Weglnge Ant-Net", antNetLengthField);

        dijkstraLengthField = new JTextField("-");
        dijkstraLengthField.setHorizontalAlignment(JTextField.RIGHT);
        dijkstraLengthField.setEditable(false);
        items.addItem("Weglnge Dijkstra", dijkstraLengthField);

        relativeLengthField = new JTextField("-");
        relativeLengthField.setHorizontalAlignment(JTextField.RIGHT);
        relativeLengthField.setEditable(false);
        items.addItem("Verhltnis Ant-Net / Dijkstra", relativeLengthField);




        delaySlider = createSlider();
        items.addItem("Verzgerung", delaySlider);
    }
    
    /**
     * Erzeugt einen Slider zur Einstellung der Animationsgeschwindigkeit.
     * Der Bereich der einstellbaren Zeitverzoegerung liegt zwischen 0
     * und 4 Sekunden (Default = 2).
     * @return Slider-Objekt.
     */
    private JSlider createSlider() {
        final JSlider slider = new JSlider(JSlider.HORIZONTAL,
                                          0, 7, 4);
        slider.setMajorTickSpacing(1);
        slider.setPaintTicks(true);
        slider.addChangeListener(new ChangeListener() {
            public void stateChanged(ChangeEvent e) {
                if (!slider.getValueIsAdjusting()) {
                    setDelayTime();
                }
            }
        });
        return slider;
    }
    
    /**
     * Die fest eingestellten Stufen der Verzoegerungszeit.
     * (alle Werte sind in Millisekunden)
     */
    private static final long[] DELAY_TIMES = 
        {0, 1, 100, 200, 400, 700, 1500, 3000};
    
    /**
     * Uebermittelt die in delaySlider eingestellte
     * Verzoegerungszeit an den Loesungsalgorithmus
     */
    private void setDelayTime() {
        long delay = DELAY_TIMES[(int)delaySlider.getValue()];
        solver.setDelay(delay);     
    }
    
    /**
     * Startet/stoppt die Suche.
     * Das Verhalten entspricht einem "Toggle".
     * Beim Start werden die Werte des Panels uebernommen.
     * Da die Suche in einem eigenen Thread laeuft, kehrt die
     * Methode sofort zurueck.
     */
    //todo: die alte Anwendung
    private void startApplication() {
        if (solver.isRunning()) {
            setButtonToStart();
            solver.stop();
        }
        else {
            //setButtonToStop();
            String startName = startSelection.getSelectedItem().toString();
            String goalName = goalSelection.getSelectedItem().toString();
            //int index = algoSelection.getSelectedIndex();
            setDelayTime();
            solver.start(
                algorithms[/*index*/0],
                graph.findNode(startName),
                graph.findNode(goalName));
        }
    }

    /**
     * Startet/stoppt die Ameisen-Suche.
     * Das Verhalten entspricht einem "Toggle".
     * Beim Start werden zufllige Werte erzeugt
     * Da die Suche in einem eigenen Thread laeuft, kehrt die
     * Methode sofort zurueck.
     */
    private void startAntApplication() {
        if (solver.isRunning()) {
            setButtonToStart();
            solver.stop();
        }
        else {
            //setButtonToStop();
            setDelayTime();

                int randomStart = (int)(Math.random() * graph.getNodeNames().length);
                //int randomGoal = (int)(Math.random() * graph.getNodeNames().length);
                //String[] goals = {"Hamburg","Aachen","Mnchen","Berlin","Stuttgart"};
                //int randomStart = (int)(Math.random() * goals.length);
                //int randomGoal = (int)(Math.random() * goals.length);
                //startName = goals[randomStart];//"Flensburg";//startSelection.getSelectedItem().toString();

                startName = graph.getNodeNames()[randomStart];
                goalName = "Frankfurt";//goals[randomGoal];//graph.getNodeNames()[randomGoal];//"Mnchen";//goalSelection.getSelectedItem().toString();

                //int index = algoSelection.getSelectedIndex();
                //setDelayTime();

                if (dijkstraCheckbox.isSelected()){
                    //die nchste Suche wird eine Bestwegsuche
                    antToggle = false;
                }
                solver.setExplorationProbability(Double.parseDouble(explorationProbability.getText()));
                solver.setTimeToLive(Integer.parseInt(timeToLive.getText()));
                solver.start(
                    SearchStrategy.ANT_NET,
                    graph.findNode(startName),
                    graph.findNode(goalName));

        }
    }

    /**
     * Startet/stoppt die Bestweg-Suche.
     * Das Verhalten entspricht einem "Toggle".
     * Beim Start werden die selben Start- und Zieldaten wie bei der Ameisensuche benutzt
     * Da die Suche in einem eigenen Thread laeuft, kehrt die
     * Methode sofort zurueck.
     */
    private void startPathLengthApplication() {
        if (solver.isRunning()) {
            setButtonToStart();
            solver.stop();
        }
        else {
            //setButtonToStop();
            setDelayTime();

                //Die nchste Suche soll wieder eine Ameisensuche werden
                antToggle = true;
                solver.start(
                    SearchStrategy.PATH_LENGTH,
                    graph.findNode(startName),
                    graph.findNode(goalName));

        }
    }

    /* (non-Javadoc)
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    public void update(Observable o, Object arg) {
        Path path = (Path)arg;

        if (!solver.isRunning() && pleaseContinue) {
            // Aenderungen der GUI (auer JTextField)
            // sollten nur im Event-Thread erfolgen !

            //Knoten entfernen
            if (removeCheckbox.isSelected()){
                    Node node = graph.findNode("Giessen");
                    if (null != node){
                        System.out.println("***************************************");
                        graph.removeNode(node);
                    }
            }

            //neue Ameisensuche
            if (antToggle){


                //Es ist gerade eine Bestwegsuche gelaufen
                if (dijkstraCheckbox.isSelected() && antNetLength!=0.0){
                    dijkstraLengthField.setText(lengthFormat.format(path.getPathLength()) + " km");
                    dijkstraLength = path.getPathLength();
                    double d = antNetLength / dijkstraLength;
                    relativeLengthField.setText(lengthFormat.format(d));
                }

                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        //setButtonToStart();
                        startAntApplication();
                    }
                });
            }
            else {
                //Es ist gerade eine Ant-Net-Suche gelaufen
                if (dijkstraCheckbox.isSelected() && path.isComplete()){
                    antNetLengthField.setText(lengthFormat.format(path.getPathLength()) + " km");
                    antNetLength = path.getPathLength();
                    dijkstraLengthField.setText("-");
                    relativeLengthField.setText("-");


                }
                if (!path.isComplete()){
                    antNetLengthField.setText("-");
                    antNetLength=0.0;
                }
                final boolean complete = path.isComplete();
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        //nur wenn die Ameisensuche erfolgreich war wird eine Bestwegsuche als Vergleich gestartet
                        if (complete){
                            startPathLengthApplication();
                        }
                        else{
                            startAntApplication();
                        }
                    }
                });
            }
        }
    }
    
    /**
     * Definiert das Aussehen des Startknopfes.
     */
    private void setButtonToStart() {
        startButton.setFont(new Font("SansSerif", Font.BOLD, 18));
        startButton.setBackground(BUTTON_COLOR);
        startButton.setForeground(START_COLOR);
        startButton.setText("start");
    }
    
    /**
     * Definiert das Aussehen des Stoppknopfes.
     */
    private void setButtonToStop() {
        startButton.setFont(new Font("SansSerif", Font.BOLD, 18));
        startButton.setBackground(BUTTON_COLOR);
        startButton.setForeground(STOP_COLOR);
        startButton.setText("stop");
    }
}
