/*
 * Roadmap.java
 * Created on 22.04.2003
 */
package search.gui;

import search.graph.*;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.util.Iterator;
import java.util.List;
import java.util.Observable;
import java.util.Observer;

import javax.swing.JPanel;

/**
 * Stellt den aktuellen Zustand der Suche graphisch dar.
 * @author Erich Ehses.
 */
public final class Roadmap extends JPanel implements Observer {
    
    /*
     * Konstanten fr das Aussehen der Graphik
     */
    private static final int NODE_RADIUS = 8;    
    private static final Color GRAPH_COLOR = Color.black;
    private static final Color PATH_COLOR = Color.red;
    private static final Color CLOSED_COLOR = Color.green;
    private static final Color BG_COLOR = Color.white;
    private static final int INSET = 25;

    /*
     * Modelldaten.
     */    
    private Graph graph;
    private List currentPath;
    private Node goalNode;

    /*
     * aktueller (Hintergrund) Kontext fr die Graphik.
     */
    private Image img = null;
    private Graphics g0;

    /*
     * Dimension der graphischen Darstellung.
     */
    private int _width;
    private int _height;
    
    /*
     * Skalierungsgroessen.
     */
    private double sX;
    private double sY;
    private double offX;
    private double offY;
    
    /**
     * Konstruktor.
     * Breite und Hoehe fuehren zu einer automatischen Skalierung.
     * @param graph der Graph in dem die Suche stattfindet.
     * @param width die Breite der Darstellung.
     * @param height die Hoehe der Darstellung.
     */
    public Roadmap(Graph graph, int width, int height) {
        this.graph = graph;
        _width = width;
        _height = height;
        findScale();
        setPreferredSize(new Dimension(width, height));
        //setDoubleBuffered(false);
    }

    /* (non-Javadoc)
     * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
     */
    public synchronized void paintComponent(Graphics g) {
        // Zeichne den Graphen (mit eventueller vorheriger Pfadinformation)
        // als Hintergrund.
        drawGraph(g);
        // Markiere den Pfad und den Zielknoten:
        drawPath(g);
        if (goalNode != null) {
            int x = scaleX(goalNode.getX());
            int y = scaleY(goalNode.getY());
            int r = NODE_RADIUS + 4;
            g.fillArc(x - r / 2, y - r / 2, r, r, 0, 360);
        }
    }
    
    /* (non-Javadoc)
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    public synchronized void update(Observable o, Object arg) {
        final Path p = (Path)arg;
        currentPath = p.getNodes();
        if (currentPath.size() <= 1) img = null;
        goalNode = p.getGoalNode();
        repaint();
    }    

    /**
     * Bestimmt die Skalierungparameter.
     */
    private void findScale() {
        Iterator iter = graph.nodeIterator();
        Node first = (Node)iter.next();
        double minX = first.getX();
        double maxX = minX;
        double minY = first.getY();
        double maxY = minY;
        while (iter.hasNext()) {
            Node node = (Node)iter.next();
            double x = node.getX();
            double y = node.getY();
            if (x < minX) minX = x;
            else if (x > maxX) maxX = x;
            if (y < minY) minY = y;
            else if (y > maxY)  maxY = y;
        }
        double dx = (maxX - minX);
        double dy = (minY - maxY);
        int w = _width - 2*INSET;
        int h = _height - 2*INSET;
        sX = w / dx;
        sY = h / dy;
        offX = - w * minX / dx + INSET;
        offY = -h * maxY / dy + INSET; 
    }
    
    /**
     * Transformiert x-Koordinaten auf den Bildschirm.
     * @param x x-Koordinate.
     * @return Pixelindex in x-Richtung.
     */
    private int scaleX(double x) {
        return (int)(sX*x+offX);
    }
    
    /**
     * Transformiert y-Koordinaten auf den Bildschirm.
     * @param y y-Koordinate.
     * @return Pixelindex in y-Richtung.
     */
    private int scaleY(double y) {
        return (int)(sY*y+offY);
    }
    
    /**
     * Stellt den Graphen und eventuell vorher aufgezeichnete
     * Weginformation dar.
     * Bei der Implementierung werden das <code>Image</code> Objekt
     * <code>img</code> und die damit verbundene Graphik
     * <code>g0</code> benutzt.
     * @param g aktueller Graphics-Kontext.
     */
    private void drawGraph(Graphics g) {
        if (img == null) {
            img = createImage(_width, _height);
            g0 = img.getGraphics();
            g0.setColor(BG_COLOR);
            g0.fillRect(0, 0, _width, _height);
            g0.setColor(GRAPH_COLOR);
            for (Iterator iter = graph.edgeIterator(); iter.hasNext();) {

                Teilstrecke edge = (Teilstrecke)iter.next();
                Node start = edge.getNode(0);
                Node end = edge.getNode(1);
                // hier wird die Farbe nach Teilstreckentyp verndert
                g0.setColor(edge.getColor());

                drawLine(g0,
                    start.getX(), start.getY(), end.getX(), end.getY());
                g0.setColor(GRAPH_COLOR);
            }
            for (Iterator iter = graph.nodeIterator(); iter.hasNext();) {
                Node node = (Node)iter.next();
                fillCircle(g0, scaleX(node.getX()), scaleY(node.getY()));
            }
        }
        g.drawImage(img, 0, 0, _width, _height, null);
    }
    
    /**
     * Stellt den aktuellen Weg dar. Der Weg wird gleichzeitig in
     * das Bildobjekt <code>img</code> zwecks spaeterer Darstellung
     * uebernommen.
     * @param g aktueller Graphics-Kontext.
     */
    private void drawPath(Graphics g) {
        if (currentPath != null) {
            g.setColor(PATH_COLOR);
            g0.setColor(CLOSED_COLOR);
            Iterator iter = currentPath.iterator();
            Node first = (Node)iter.next();
            if (null != first){
                int lastX = scaleX(first.getX());
                int lastY = scaleY(first.getY());
                fillCircle(g, scaleX(first.getX()), scaleY(first.getY()));
                fillCircle(g0, scaleX(first.getX()), scaleY(first.getY()));
                while (iter.hasNext()) {
                    Node node = (Node)iter.next();
                    int nextX = scaleX(node.getX());
                    int nextY = scaleY(node.getY());
                    g.drawLine(lastX, lastY, nextX, nextY);
                    g0.drawLine(lastX, lastY, nextX, nextY);
                    fillCircle(g, nextX, nextY);
                    fillCircle(g0, nextX, nextY);
                    lastX = nextX;
                    lastY = nextY;
                }
            }
        }
    }

    /**
     * Zeichne eine Wegkante.
     * @param g aktueller Graphics Kontext.
     * @param x0 Anfangspunkt.
     * @param y0 Anfangspunkt.
     * @param x1 Endpunkt.
     * @param y1 Endpunkt.
     */
    private void drawLine(Graphics g,
        double x0, double y0, double x1, double y1)
    {
        int ix0 = scaleX(x0);
        int iy0 = scaleY(y0);
        int ix1 = scaleX(x1);
        int iy1 = scaleY(y1);
        g.drawLine(ix0, iy0, ix1, iy1);             
    }

    /**
     * Zeichne einen Knoten (Ort).
     * @param g aktueller Graphics Kontext.
     * @param x x-Koordinate (Pixel).
     * @param y y-Koordinate (Pixel).
     */
    private void fillCircle(Graphics g, int x, int y) {
        g.fillArc(
            x - NODE_RADIUS / 2, y - NODE_RADIUS / 2,
            NODE_RADIUS, NODE_RADIUS, 0, 360);
    }
}
