/*
 * File SharedListTree.java.
 * Created on 27.04.2003
 */
package search.util;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

/**
 * Diese Klasse realisert eine unidirektionale Liste. Als einzige Vernderung
 * ist erlaubt, die Liste (am vorderen Ende) zu erweitern. Dadurch ist es
 * mglich, die Liste so zu optimieren, dass sich mehrere Listen die Daten
 * ab einer bestimmten Position teilen.
 * <p>
 * Alle Operationen der Schnittstelle <code>List</code> die einen wahlfreien
 * Zugriff auf einzelne Listenelemente erlauben und auch die Iteration
 * ber einen <code>ListIterator</code> sind sehr ineffizient.
 * <p>
 * Zugriffe auf das erste Element und die Iteration ber einen einfachen
 * <code>Iterator</code> sind effizient.
 * <p>
 * Zustzlich zu den Operationen der Schnittstelle <code>List</code>
 * sind Operationen fr die beim Verknpfen mehrere Listen entstehende
 * (in umgekehrte Richtung zeigende) Baumstruktur implementiert:
 * <ul>
 * <li>Die Klasse <code>TreeNode</code> beschreibt einen Baumknoten.
 *     </li>
 * <li>Die Methode <code>getRoot</code> aus der Klasse
 *     <code>SharedListTree</code> liefert die Wurzel des Baums.</li>
 * <li>Die Methode <code>getChildren</code> liefert fr einen Baumknoten
 *     die Liste seiner Kindknoten.</li>
 * <li>Die Methode <code>getValue</code> liefert den Inhalt eines Knotens.</li>
 * </ul>
 * <p>
 * Eine Konsequenz der Datenstruktur ist, dass alle erzeugten zusammenhngenden
 * Listenstrukturen erst dann in den Freispeicher recycled werden, wenn wenn
 * keine der Listen mehr existiert.
 * 
 * Die Klasse <code>SharedList</code> ist eine speichereffizientere
 * Implementierung ohne Baumstruktur.
 * 
 * @see SharedList
 *  
 * @author Erich Ehses
 */
public final class SharedListTree extends AbstractList {
    
    /*
     * Die beiden Variablen first und root referieren den Anfang und
     * das Ende der Liste. Der Name root wurde gewhlt, weil es bei der
     * umgekehrten Traversierung um einen Baum geht.
     */
    private Entry first = null;
    private Entry root = null;

    /**
     * Erzeugt eine leere Liste.
     */
    public SharedListTree() {}

    /**
     * Die neue Liste teilt sich mit der alten Liste die Daten.
     * @param oldList bereits bestehende Liste
     */
    public SharedListTree(SharedListTree oldList) {
        modCount = oldList.modCount;
        first = oldList.first;
        root = oldList.root;
    }

    /**
     * Die neue Liste ist eine Kopie der <code>Collection c</code>.
     * Wenn das Argument eine <code>SharedList</code> ist, werden die
     * Listenelemente nicht kopiert sondern gemeinsam benutzt.
     * 
     * @param c bestehendes <code>Collection</code>-Objekt.
     */
    public SharedListTree(Collection c) {
        modCount = c.size();
        if (c instanceof SharedListTree) {
            first = ((SharedListTree)c).first;
            root = ((SharedListTree)c).root;
        }
        else {
            Object[] a = c.toArray();
            int n = a.length;
            if (n != 0) {
                root = new Entry(a[a.length-1], null);
                Entry next = root;
                for (int i=a.length-2; i>=0; i--) {
                    next = new Entry(a[i], next);
                }
                first = next;
            }
        }
    }

    /* (non-Javadoc)
     * @see java.util.AbstractList#iterator()
     * @see java.util.List#iterator()
     */
    public Iterator iterator() {
        return new SharedListIterator(first);
    }

    /* (non-Javadoc)
     * @see java.util.AbstractList#size()
     * @see java.util.List#size()
     */
    public int size() {
        return modCount;
    }

    /**
     * Gibt das erste Listenelement zurueck.
     * @return erstes Element.
     */
    public Object getFirst() {
        return first.getValue();   
    }
    
    /* (non-Javadoc)
     * @see java.util.AbstractList#get(int)
     * @see java.util.List#get(int)
     */
    public Object get(int index) {
        if (index < 0 || index >= modCount)
            throw new IndexOutOfBoundsException();
        Entry p = first;
        for (int i=0; i<index; i++) p = p.next;
        return p.getValue();
    }

    /**
     * Fuegt ein Element am Anfang der Liste hinzu.
     * @param element neues Element.
     */
    public void addFirst(Object element) {
        Entry previous = first;
        first = new Entry(element, first);
        if (previous == null) {
            root = first;
        }
        modCount++;
    }
    
    /** 
     * Fuegt ein Element hinzu.
     * Nur das Hinzufuegen am Anfang der Liste ist implementiert
     * (<code>index = 0</code>).
     * @param index Position an der hinzugefuegt wird.
     * @param element Wert der eingefuegt wird.
     * 
     * @see java.util.AbstractList#add(int, java.lang.Object)
     * @see java.util.List#add(int, java.lang.Object)
     */
    public void add(int index, Object element) {
        if (index == 0) {
            addFirst(element);
        }
        else
            super.add(index, element);
    }
    
    /**
     * Ermittle die Wurzel des Baumes der
     * Rckwrtsverkettung.
     * @return Wurzel.
     */
    public TreeNode getRoot() {
        return root;
    }

    /**
     * Die Klasse stellt ein Element der Liste dar.
     * Die ffentlichen Methoden erlauben den Zugriff aber keine
     * Vernderung.
     */
    private static class Entry implements TreeNode{
        private final Object value;
        private final Entry next;
        private final java.util.List children;
        
        private Entry(Object value, Entry next) {
            this.value = value;
            this.next = next;
            children = new ArrayList();
            if (next != null) next.children.add(this);
        }
           
        /**
         * Ermittelt die Vorgngerelemente des Knotens.
         * @return Liste der Vorgnger.
         */
        public List getChildren() {
            return children;
        }
    
        /**
         * Ermittelt den Dateninhalt des Knotens.
         * @return Inhalt.
         */
        public Object getValue() {
            return value;
        }
    }

    /**
     * Implementierung eines passenden Iterators.
     */
    private static class SharedListIterator implements Iterator {
        private static Entry p;
    
        private SharedListIterator(Entry first) {
            p = first;
        }
    
        /* (non-Javadoc)
         * @see java.util.Iterator#hasNext()
         */
        public boolean hasNext() {
            return p != null;
        }
    
        /* (non-Javadoc)
         * @see java.util.Iterator#next()
         */
        public Object next() {
            Object result = p.getValue();
            p = p.next;
            return result;
        }
    
        /* (non-Javadoc)
         * @see java.util.Iterator#remove()
         */
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }    
}
