/*
 * File PriorityQueue.java.
 * Created on 10.05.2003
 */
package search.util.example;

import java.util.Comparator;
import java.util.NoSuchElementException;

import search.util.Queue;

/**
 * Bei dieser Realisierung von <code>Queue</code> wird mittels
 * <code>get</code> jeweils das gem dem angegebenen
 * <code>Comparator</code> kleinste Element zurckgegeben.
 * <p>
 * Diese Implementierung benutzt einen effizienten
 * Heap-basierten Algorithmus: O(log n) fr <code>put</code> und
 * <code>get</code>.
 * 
 * @author Erich Ehses.
 */
public class PriorityQueue implements Queue {
    
    // Anfangsgre des Datenfelds.
    private static final int DEFAULT_SIZE = 128;
    
    // Datenfeld.
    private Object[] data;
    
    // Anzahl der Datenelemente.
    private int n = 0;
    
    // Comparator-Objekt zur Festlegung der Reihenfolge.
    private Comparator cmp;

    /**
     * Erzeugt eine Priorittswarteschlange.
     * @param cmp legt die Reihenfolge fest.
     */
    public PriorityQueue(Comparator cmp) {
        this(cmp, DEFAULT_SIZE);
    }
    
    /**
     * Erzeugt eine Priorittswarteschlange mit der Anfanfangsgre
     * <code>initialSize</code>.
     * @param cmp legt die Reihenfolge fest.
     * @param initialSize Anfangsgre.
     */
    public PriorityQueue(Comparator cmp, int initialSize) {
        if (initialSize <= 0)
            throw new IllegalArgumentException("initialSize = "+initialSize);
        data = new Object[initialSize];
        this.cmp = cmp;
    }
    
    /* (non-Javadoc)
     * @see util.Queue#put(java.lang.Object)
     */
    public void put(Object obj) {
        if (n == data.length) {
            Object[] newArray = new Object[2*n];
            System.arraycopy(data, 0, newArray, 0, n);
            data = newArray;
        }
        data[n++] = obj;
        swim(n-1);
    }

    /* (non-Javadoc)
     * @see util.Queue#get()
     */
    public Object get() {
        if (this.isEmpty())
            throw new NoSuchElementException("Queue is empty");
        Object result = data[0];
        data[0] = data[--n];
        data[n] = null;
        sink(0);
        return result;
    }

    /* (non-Javadoc)
     * @see util.Queue#isEmpty()
     */
    public boolean isEmpty() {
        return n == 0;
    }

    /* (non-Javadoc)
     * @see util.Queue#clear()
     */
    public void clear() {
        for (int i=0; i<n; i++)
            data[i] = null;
        n = 0;
    }
    
    /**
     * Promoviert einen Knoten soweit, bis er auf ein Elter mit
     * besseren Eigenschaften stt.
     * (Der Knoten schwimmt nach oben, bis er seinen Platz gefunden hat.)
     * @param child Index des Kindknotens.
     */
    private void swim(int child) {
        int parent = (child-1) / 2;
        while (child > 0 && isBetter(child, parent)) {
            swap(child, parent);
            child = parent;
            parent = (child-1) / 2;
        }
    }
    
    /**
     * Degradiert einen Knoten soweit bis er auf Kinder mit
     * schlechteren Eigenschaften stt.
     * (Der Knoten sinkt nach unten, bis er seinen Platz gefunden hat.)
     * @param parent Index des Elternknotens.
     */
    private void sink(int parent) {
        int child = 2*parent+1;
        while (child < n) {
            if (child < n-1 && isBetter(child+1, child)) child++;
            if (isBetter(parent, child)) return;
            swap(parent, child);
            parent = child;
            child = 2*parent+1;
        }
    }
    
    /**
     * Vertausche zwei Knoten.
     * @param i Erster Knoten.
     * @param j Zweiter Knoten.
     */
    private void swap(int i, int j) {
        Object temp = data[i];
        data[i] = data[j];
        data[j] = temp;
    }
    
    /**
     * Vergleiche zwei Knoten.
     * @param i Erster Knoten
     * @param j Zweiter Knoten.
     * @return <code>true</code> wenn der 1. Knoten besser als der 2. ist.
     */
    private boolean isBetter(int i, int j) {
        return cmp.compare(data[i], data[j]) < 0;
    }

}
